#
#  Copyright (c) 2014, Facebook, Inc.
#  All rights reserved.
#
#  This source code is licensed under the BSD-style license found in the
#  LICENSE file in the root directory of this source tree. An additional grant
#  of patent rights can be found in the PATENTS file in the same directory.
#
#
from atc_api.settings import atc_api_settings
from atc_thrift import Atcd
from thrift.transport import TSocket, TTransport
from thrift.protocol import TBinaryProtocol


def atcdClient():
    try:
        transport = TSocket.TSocket(
            atc_api_settings.ATCD_HOST,
            atc_api_settings.ATCD_PORT
        )
        transport = TTransport.TFramedTransport(transport)
        transport.open()
        protocol = TBinaryProtocol.TBinaryProtocol(transport)
        return Atcd.Client(protocol)
    except TTransport.TTransportException as e:
        print 'atcdClient: %s: %s' % (e.__class__.__name__, str(e))



#
#  Copyright (c) 2014, Facebook, Inc.
#  All rights reserved.
#
#  This source code is licensed under the BSD-style license found in the
#  LICENSE file in the root directory of this source tree. An additional grant
#  of patent rights can be found in the PATENTS file in the same directory.
#
#
from atc_api.settings import atc_api_settings
from atc_thrift.ttypes import Corruption
from atc_thrift.ttypes import Delay
from atc_thrift.ttypes import Loss
from atc_thrift.ttypes import Reorder
from atc_thrift.ttypes import Shaping
from atc_thrift.ttypes import TrafficControlledDevice
from atc_thrift.ttypes import TrafficControlSetting

from rest_framework import serializers
from thrift.Thrift import TType

import socket


def validate_ipaddr(ipaddr):
    try:
        socket.inet_aton(ipaddr)
        return True
    except socket.error:
        return False


class ThriftSerializer(serializers.Serializer):
    # Should be set by the serializer to the concrete thrift class
    # to be serialized.
    _THRIFT_CLASS = None

    # A map of renamed fields.
    # Keys in the map are the names of thrift fields. Their values
    # are the names of the serializer fields they correspond to.
    _THRIFT_RENAMED_FIELDS = {}

    def create(self, attrs):
        args = {}

        for field_tuple in self._THRIFT_CLASS.thrift_spec:
            if not field_tuple:
                continue

            _, thrift_type, arg_name, _, default = field_tuple

            f_name = arg_name
            if arg_name in self._THRIFT_RENAMED_FIELDS:
                f_name = self._THRIFT_RENAMED_FIELDS[arg_name]

            serializer = self.fields[f_name]

            if f_name not in attrs:
                args[arg_name] = default
                continue

            if thrift_type == TType.STRUCT:
                args[arg_name] = serializer.create(attrs[f_name])
            else:
                # Primitive
                args[arg_name] = attrs[f_name]

        return self._THRIFT_CLASS(**args)


class BaseShapingSettingSerializer(ThriftSerializer):
    percentage = serializers.FloatField(default=0)
    correlation = serializers.FloatField(default=0)


class DelaySerializer(ThriftSerializer):
    _THRIFT_CLASS = Delay

    delay = serializers.IntegerField(default=0)
    jitter = serializers.IntegerField(default=0)
    correlation = serializers.FloatField(default=0)


class LossSerializer(BaseShapingSettingSerializer):
    _THRIFT_CLASS = Loss


class CorruptionSerializer(BaseShapingSettingSerializer):
    _THRIFT_CLASS = Corruption


class ReorderSerializer(BaseShapingSettingSerializer):
    _THRIFT_CLASS = Reorder

    gap = serializers.IntegerField(default=0)


class IptablesOptionsField(serializers.Field):

    def to_representation(self, data):
        if isinstance(data, list):
            return data
        else:
            msg = self.error_messages['invalid']
            raise serializers.ValidationError(msg)

    def to_internal_value(self, obj):
        if obj:
            return obj
        else:
            return []


class ShapingSerializer(ThriftSerializer):
    _THRIFT_CLASS = Shaping

    rate = serializers.IntegerField(default=0, allow_null=True, required=False)
    loss = LossSerializer(default=None, allow_null=True, required=False)
    delay = DelaySerializer(default=None, allow_null=True, required=False)
    corruption = CorruptionSerializer(
        default=None, allow_null=True, required=False)
    reorder = ReorderSerializer(default=None, allow_null=True, required=False)
    iptables_options = IptablesOptionsField(
        default=None, allow_null=True, required=False)


class SettingSerializer(ThriftSerializer):
    _THRIFT_CLASS = TrafficControlSetting

    down = ShapingSerializer()
    up = ShapingSerializer()


class DeviceSerializer(ThriftSerializer):
    _THRIFT_CLASS = TrafficControlledDevice
    _THRIFT_RENAMED_FIELDS = {
        'controllingIP': 'client',
        'controlledIP': 'address'
    }

    address = serializers.CharField(
        max_length=16,
        allow_blank=True,
        allow_null=True,
        default=None,
        required=False
    )
    client = serializers.CharField(
        max_length=16,
        allow_blank=True,
        allow_null=True,
        default=None,
        required=False
    )

    def validate_address(self, value):
        # 'address' is optional, if not specified, we default to the
        # querying IP
        # `address` can be specified in 2 places: the URL or within the payload
        # The payload has priority and will be accessible through `value`
        # The value passed in the URL is accessible through the context

        if value is None or (isinstance(value, str) and len(value) == 0):
            if self.context.get('address'):
                value = self.context['address']
            else:
                value = self._get_client_ip()
        if not validate_ipaddr(value):
            raise serializers.ValidationError("Invalid IP address")
        return value

    def validate_client(self, value):
        # 'client' should not be provided by the payload.
        # It should always be the client IP as we get it from _get_client_ip()
        # This is merely here so we can use the serializer.
        return self._get_client_ip()

    def _get_client_ip(self):
        '''Return the real IP of a client even when using a proxy'''
        request = self.context['request']
        if 'HTTP_X_REAL_IP' in request.META:
            if request.META['REMOTE_ADDR'] not in atc_api_settings.PROXY_IPS:
                raise ValueError('HTTP_X_REAL_IP set by non-proxy')
            return request.META['HTTP_X_REAL_IP']
        else:
            return request.META['REMOTE_ADDR']



#!/usr/bin/env python
#
#  Copyright (c) 2014, Facebook, Inc.
#  All rights reserved.
#
#  This source code is licensed under the BSD-style license found in the
#  LICENSE file in the root directory of this source tree. An additional grant
#  of patent rights can be found in the PATENTS file in the same directory.
#
#
import os
import re
import sys

from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}

version = get_version('atc_api')

if sys.argv[-1] == 'publish':
    if os.system("pip freeze | grep wheel"):
        print("wheel not installed.\nUse `pip install wheel`.\nExiting.")
        sys.exit()
    if os.system("pip freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload dist/*")
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    sys.exit()

setup(
    name='django-atc-api',
    version=version,
    description='REST API for ATC',
    author='Emmanuel Bretelle',
    author_email='chantra@fb.com',
    url='https://github.com/facebook/augmented-traffic-control',
    packages=get_packages('atc_api'),
    package_data=get_package_data('atc_api'),
    classifiers=['Programming Language :: Python', ],
    long_description=README,
    install_requires=['atc_thrift', 'djangorestframework']
)



#
#  Copyright (c) 2014, Facebook, Inc.
#  All rights reserved.
#
#  This source code is licensed under the BSD-style license found in the
#  LICENSE file in the root directory of this source tree. An additional grant
#  of patent rights can be found in the PATENTS file in the same directory.
#
#
from atc_api.atcd_client import atcdClient
from atc_api.serializers import SettingSerializer, DeviceSerializer
from atc_api.settings import atc_api_settings
from atc_thrift.ttypes import TrafficControlException, TrafficControl
from atc_thrift.ttypes import TrafficControlledDevice, AccessToken

from functools import wraps
from rest_framework.exceptions import APIException
from rest_framework.exceptions import ParseError
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework import status


class BadGateway(APIException):
    status_code = 502
    default_detail = 'Could not connect to ATC gateway.'


def serviced(method):
    '''
    A decorator to check if the service is available or not.
    Raise a BadGateway exception on failure to connect to the atc gateway
    '''
    @wraps(method)
    def decorator(cls, request, *args, **kwargs):
        service = atcdClient()
        if service is None:
            raise BadGateway()
        return method(cls, request, service, *args, **kwargs)
    return decorator


class AtcApi(APIView):
    '''
    If `address` is not provided, we default to the client IP or forwarded IP
    '''

    @serviced
    def get(self, request, service, address=None, format=None):
        ''''
        Get the current shaping for an IP. If address is None, defaults to
        the client IP
        @return the current shaping applied or 404 if the IP is not being
        shaped
        '''
        device_serializer = DeviceSerializer(
            data=request.DATA,
            context={'request': request, 'address': address},
        )
        if not device_serializer.is_valid():
            raise ParseError(detail=device_serializer.errors)

        dev = device_serializer.save()
        try:
            tc = service.getCurrentShaping(dev)
        except TrafficControlException as e:
            return Response(
                {'detail': e.message},
                status=status.HTTP_404_NOT_FOUND,
            )

        serializer = SettingSerializer(tc.settings)
        return Response(
            serializer.data,
            status=status.HTTP_200_OK
        )

    @serviced
    def post(self, request, service, address=None, format=None):
        ''''
        Set shaping for an IP. If address is None, defaults to
        the client IP
        @return the profile that was set on success
        '''
        setting_serializer = SettingSerializer(data=request.DATA)
        device_serializer = DeviceSerializer(
            data=request.DATA,
            context={'request': request, 'address': address},
        )
        if not setting_serializer.is_valid():
            raise ParseError(detail=setting_serializer.errors)

        if not device_serializer.is_valid():
            raise ParseError(detail=device_serializer.errors)

        setting = setting_serializer.save()
        device = device_serializer.save()

        tc = TrafficControl(
            device=device,
            settings=setting,
            timeout=atc_api_settings.DEFAULT_TC_TIMEOUT,
        )

        try:
            tcrc = service.startShaping(tc)
        except TrafficControlException as e:
            return Response(e.message, status=status.HTTP_401_UNAUTHORIZED)
        result = {'result': tcrc.code, 'message': tcrc.message}
        if tcrc.code:
            raise ParseError(detail=result)

        return Response(
            setting_serializer.data,
            status=status.HTTP_201_CREATED
        )

    @serviced
    def delete(self, request, service, address=None, format=None):
        '''
        Delete the shaping for an IP, if no IP is specified, default to the
        client IP
        '''
        device_serializer = DeviceSerializer(
            data=request.DATA,
            context={'request': request, 'address': address},
        )
        if not device_serializer.is_valid():
            return Response(
                device_serializer.errors,
                status=status.HTTP_400_BAD_REQUEST,
            )

        device = device_serializer.save()

        try:
            tcrc = service.stopShaping(device)
        except TrafficControlException as e:
            return Response(e.message, status=status.HTTP_401_UNAUTHORIZED)

        result = {'result': tcrc.code, 'message': tcrc.message}
        if tcrc.code:
            raise ParseError(detail=result)
        return Response(status=status.HTTP_204_NO_CONTENT)


class AuthApi(APIView):

    @serviced
    def get(self, request, service, address=None):
        '''
        Returns the addresses that the provided address is allowed to shape.
        '''
        if address is None:
            address = _get_client_ip(request)

        controlled_ips = []

        for addr in service.getDevicesControlledBy(address):
            if addr is None:
                break
            controlled_ips.append({
                'controlled_ip': addr.device.controlledIP,
                'valid_until': addr.timeout,
            })

        data = {
            'address': address,
            'controlled_ips': controlled_ips,
        }
        return Response(data, status=status.HTTP_200_OK)

    @serviced
    def post(self, request, service, address=None):
        '''
        Authorizes one address to shape another address,
        based on the provided auth token.
        '''
        if address is None:
            return Response(
                {'details': 'no address provided'},
                status=status.HTTP_400_BAD_REQUEST
                )
        controlled_ip = address

        controlling_ip = _get_client_ip(request)

        if 'token' not in request.data:
            token = None
        else:
            token = AccessToken(token=request.data['token'])

        dev = TrafficControlledDevice(
            controlledIP=controlled_ip,
            controllingIP=controlling_ip
            )

        worked = service.requestRemoteControl(dev, token)

        if not worked:
            return Response(
                {'details': 'invalid token provided'},
                status=status.HTTP_401_UNAUTHORIZED,
                )

        print 'Worked:', worked

        data = {
            'controlling_ip': controlling_ip,
            'controlled_ip': controlled_ip,
        }

        return Response(data, status=status.HTTP_200_OK)


class TokenApi(APIView):

    @serviced
    def get(self, request, service):
        '''
        Returns the current authorization token for the provided address.
        '''
        # default duration...
        # 3 days in seconds
        duration = 3 * 24 * 60 * 60

        if 'duration' in request.query_params:
            duration = int(request.query_params['duration'])

        address = _get_client_ip(request)

        stuff = service.requestToken(address, duration)

        data = {
            'token': stuff.token,
            'interval': stuff.interval,
            'valid_until': stuff.valid_until,
            'address': address,
        }

        return Response(data, status=status.HTTP_200_OK)


def _get_client_ip(request):
    '''Return the real IP of a client even when using a proxy'''

    if 'HTTP_X_REAL_IP' in request.META:
        if request.META['REMOTE_ADDR'] not in atc_api_settings.PROXY_IPS:
            raise ValueError('HTTP_X_REAL_IP set by non-proxy')
        return request.META['HTTP_X_REAL_IP']
    else:
        return request.META['REMOTE_ADDR']



#
#  Copyright (c) 2014, Facebook, Inc.
#  All rights reserved.
#
#  This source code is licensed under the BSD-style license found in the
#  LICENSE file in the root directory of this source tree. An additional grant
#  of patent rights can be found in the PATENTS file in the same directory.
#
#
from django.conf.urls import patterns
from django.conf.urls import url
from atc_api.views import AtcApi, AuthApi, TokenApi

urlpatterns = patterns(
    '',
    url('^shape/$', AtcApi.as_view()),
    url('^shape/'
        '(?P<address>[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})/$',
        AtcApi.as_view()
        ),
    url('^auth/$', AuthApi.as_view()),
    url('^auth/'
        '(?P<address>[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})/$',
        AuthApi.as_view()),
    url('^token/$', TokenApi.as_view()),
)



#
#  Copyright (c) 2014, Facebook, Inc.
#  All rights reserved.
#
#  This source code is licensed under the BSD-style license found in the
#  LICENSE file in the root directory of this source tree. An additional grant
#  of patent rights can be found in the PATENTS file in the same directory.
#
#
'''
Settings for ATC_API are all namespaced in the ATC_API setting.
For example your project's `settings.py` file might look like this:

ATC_API = {
    'ATCD_HOST': 'localhost',
    'ATCD_PORT': 9090,
    'DEFAULT_TC_TIMEOUT': 24 * 60 * 60,
    'PROXY_IPS': ['127.0.0.1'],
}

This module provides the `atc_api_settings` object, that is used to access
ATC_API settings. It first check for user settings and then fall back on the
defaults.
'''

from django.conf import settings

USER_SETTINGS = getattr(settings, 'ATC_API', None)

DEFAULTS = {
    'ATCD_HOST': 'localhost',
    'ATCD_PORT': 9090,
    # Default timeout is a day in seconds
    'DEFAULT_TC_TIMEOUT': 24 * 60 * 60,
    'PROXY_IPS': ['127.0.0.1'],
}


class APISettings(object):

    def __init__(self, user_settings=None, defaults=None):
        self.__user_settings = user_settings or {}
        self.__defaults = defaults or {}

    def __getattr__(self, attr):
        if attr not in self.__defaults.keys():
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            # Check if user have set that key.
            val = self.__user_settings[attr]
        except KeyError:
            # Use defaults otherwise.
            val = self.__defaults[attr]

        return val


atc_api_settings = APISettings(USER_SETTINGS, DEFAULTS)
